from __future__ import annotations

import itertools

import jijmodeling as jm
import numpy as np

from jijzept.client import JijZeptClient
from jijzept.sampler.base_sampler import ParameterSearchParameters, sample_model


def qubo_problem(
    qubo: dict[tuple[int, int], float] | dict[tuple[str, str], float], constant: float
) -> tuple[
    jm.Problem, dict[str, np.ndarray | list], dict[int, str], dict[str, jm.Binary]
]:
    """Convert the coefficients of a QUBO represented as a dictionary to a JijModeling Problem object.

    Args:
    - qubo: A dictionary representing a QUBO problem, where the keys are tuples of integers or strings
           representing the indices of the decision variables, and the values are floats representing
           the coefficients of the quadratic terms of the QUBO.
    - constant: A float representing the constant energy offset in the QUBO problem.

    Returns:
    - A tuple containing:
        - The QUBO problem as a JijModeling Problem object.
        - The instance data as a dictionary containing the values of the placeholders of the problem.
        - A dictionary mapping the indices of the decision variables to their corresponding labels (if the
          indices are represented as strings).
        - A dictionary containing the decision variables as JijModeling Binary objects.
    """
    qubo_keys = list(qubo.keys())
    unique_labels: dict[str, int] = {}
    N = jm.Placeholder("N")
    key = qubo_keys[0][0]
    if isinstance(key, int):
        # decision_variables = [jm.Binary("x", shape=N)]
        pass
    elif isinstance(key, str):
        unique_labels = {
            k: i for i, k in enumerate(itertools.chain.from_iterable(qubo_keys))
        }
        qubo_keys = [[unique_labels[e0], unique_labels[e1]] for e0, e1 in qubo_keys]
    else:
        TypeError(
            f"`{key.__class__.__name__}` object is not supported as a key of QUBO."
        )

    # Create QUBO Problem
    L = jm.Placeholder("L")
    E = jm.Placeholder("E", shape=(L, 2))
    Q = jm.Placeholder("Q", shape=L)
    c = jm.Placeholder("c")
    e = jm.Element("e", (0, L))
    x = jm.Binary("x", shape=N)
    qubo_model = jm.Problem("qubo")
    qubo_model += jm.Sum(e, Q[e] * x[E[e, 0]] * x[E[e, 1]]) + c

    edges = np.array(qubo_keys, dtype=int)
    values = np.array(list(qubo.values()), dtype=float)
    instance_data = {
        "N": int(np.max(edges)),
        "L": int(values.shape[0]),
        "E": edges.tolist(),
        "Q": values.tolist(),
        "c": constant,
    }

    return qubo_model, instance_data, {i: k for k, i in unique_labels.items()}, {"x": x}


def sample_qubo(
    client: JijZeptClient,
    solver: str,
    queue_name: str,
    qubo: dict[tuple[int, int], float],
    constant: float,
    max_wait_time: int | float | None,
    sync: bool,
    **kwargs,
):
    """Sample QUBO problem using the specified solver.

    Args:
    - client: A JijZeptClient object representing the client used to submit the problem.
    - solver: A string representing the name of the solver used to solve the problem.
    - queue_name: A string representing the name of the queue to use for the problem.
    - qubo: A dictionary representing a QUBO problem, where the keys are tuples of integers representing
           the indices of the decision variables, and the values are floats representing the coefficients
           of the quadratic terms of the QUBO.
    - constant: A float representing the constant energy offset in the QUBO problem.
    - max_wait_time: An integer or float representing the maximum time to wait for the problem to complete
                     (in seconds). If None, wait indefinitely.
    - sync: A boolean representing whether to wait for the problem to complete before returning the results.
    - **kwargs: Additional arguments to pass to the sample_model function.

    Returns:
    - A SampleSet object: Stores minimum energy samples and other information.
    """
    qubo_model, instance_data, var_map, _ = qubo_problem(qubo, constant)

    parameters = ParameterSearchParameters(
        multipliers={}, mul_search=False, normalize_qubo=False
    )

    sample_set = sample_model(
        client=client,
        solver=solver,
        queue_name=queue_name,
        problem=qubo_model,
        instance_data=instance_data,
        fixed_variables={},
        parameter_search_parameters=parameters,
        max_wait_time=max_wait_time,
        sync=sync,
        **kwargs,
    )
    sample_set.set_variable_labels(var_map)
    return sample_set
